const SVGObjectType = {
  Area: 'area',
  Image: 'image',
  Text: 'text',
  Line: 'line',
  Filter: 'filter',
};

const appBase = import.meta.url.split('?')[0].split('/').slice(0, -2).join('/');

function publicUrl (url : string) {
  return url[0] === '/' ? appBase + url : appBase + '/' + url;
}

function createTextMeasurer() {
  let testSVG = document.createElementNS('http://www.w3.org/2000/svg', 'svg');
  testSVG.classList.add('text-measurer');

  testSVG.style.position = 'absolute';
  testSVG.style.visibility = 'hidden';
  testSVG.style.left = '0';
  testSVG.style.top = '0';
  testSVG.style.width = '0';
  testSVG.style.height = '0';
  testSVG.style.padding = '0';
  testSVG.style.margin = '0';

  document.body.appendChild(testSVG);

  let elem = document.createElementNS('http://www.w3.org/2000/svg', 'text');
  elem.setAttribute('line-height', '1');
  testSVG.appendChild(elem);

  return elem;
}

let textMeasurer : SVGTextElement | undefined;

function reformatHTML(str = '') {
  let elem = (new DOMParser()).parseFromString('<html><body></body></html>', 'text/html').querySelector('body');

  if (elem) {
    elem.innerHTML = str;
    str = elem.innerHTML;
    return str.replaceAll(/<(i|em)>/g, '<tspan font-style="italic" role="emphasis">')
      .replaceAll(/<\/(i|em)>/g, '</tspan>')
      .replaceAll(/<(u)>/g, '<tspan text-decoration="underline">')
      .replaceAll(/<\/(u)>/g, '</tspan>')
      .replaceAll(/<(b|strong)>/g, '<tspan font-weight="700" role="strong">')
      .replaceAll(/<\/(b|strong)>/g, '</tspan>');
  }

  return str;
}

function measureSVGText(text : string, size : string = '16pt', font : string = 'Tahoma, sans-serif', weight : number = 400, style? :string, decoration? : string) : ({ width : number, height : number } | void) {
  if (text.length) {
    if (!textMeasurer) {
      textMeasurer = createTextMeasurer();
    }

    if (size && size.length) {
      textMeasurer.setAttribute('font-size', size);
    }

    if (font && font.length) {
      textMeasurer.setAttribute('font-family', font);
    }

    if (weight) {
      textMeasurer.setAttribute('font-weight', weight.toString());
    }

    if (style) {
      textMeasurer.setAttribute('font-style', style);
    }

    if (decoration) {
      textMeasurer.setAttribute('text-decoration', decoration);
    }

    textMeasurer.innerHTML = reformatHTML(text);
    let rect = textMeasurer.getBBox();
    return { width: rect.width, height: rect.height };
  }
}

class NumberPair {
  start : number = 0;
  end : number = 1;

  constructor(start : number = 0, end : number = 0) {
    this.start = start;
    this.end = end;
  }
};

class Point {
  x = 0;
  y = 0;

  constructor(x : number = 0, y : number = 0) {
    this.x = x;
    this.y = y;
  }

  public toString(separator : string = ',') {
    return this.x + separator + this.y;
  }
};

class AreaSVGObject implements SVGObjectInterface {
  type : string = SVGObjectType.Area;
  category : string | undefined = undefined;
  position : Point = new Point();
  layer : number = 0;
  popup = {
    title: '',
    text: '',
    imageUrl: '',
    imageWidth : 0,
    imageHeight : 0,
    link: '',
  };

  points : Point[] = [
    new Point(-15, -15),
    new Point(15, -15),
    new Point(15, 15),
    new Point(-15, 15),
  ];
  fill = '#05a04580';
  stroke = '#05a045FF';
  border = true;
  mask : string | undefined = undefined;

  text : string | undefined = '';
  textColor = '#000000FF';
  fontSize : number = 16;
  fontWeight : number = 400;
  fontStyle : string | null = null;

  constructor(o : object|undefined = undefined) {
    Object.assign(this, o);

    if (typeof o === 'object') {
      if ('position' in o && (typeof o.position === 'object') && o.position !== null && 'x' in o.position && 'y' in o.position && typeof o.position.x === 'number' && typeof o.position.y === 'number') {
        this.position = new Point(o.position.x, o.position.y);
      }

      this.fontWeight = this.fontWeight || 400;
      this.fontStyle = this.fontStyle || null;
      this.points = this.points.map((o : Point) => new Point(o.x, o.y));
      this.layer = this.layer || 0;
    }
  }

  public polygonPointString() : string {
    return this.points.map(p => p.toString()).join(' ');
  }

  public clipPathString() : string {
    return 'path(\'M ' + this.points.map(p => p.toString(' ')).join(' L ') + ' z\')';
  }

  public getLinePointIndexes() : NumberPair[] {
    return Object.keys(this.points).map(Number).map(i => {
      return new NumberPair(i % this.points.length, (i + 1) % this.points.length);
    });
  }

  public getTransformOrigin() {
    let ret = new Point();

    if (this.points.length > 0) {
      this.points.forEach(p => {
        ret.x += p.x;
        ret.y += p.y;
      });

      ret.x /= this.points.length;
      ret.y /= this.points.length;
    }

    return ret;
  }

  public getTransform() {
    return `translate(${this.position.x},${this.position.y})`;
  }

  public hasPopupImage() {
    return Boolean(this.popup.imageUrl && this.popup.imageUrl.length && this.popup.imageHeight && this.popup.imageWidth);
  }

  public hasPopupTitle() {
    return Boolean(this.popup.title.length);
  }

  public hasPopupText() {
    return Boolean(this.popup.text.length);
  }

  public hasPopupLink() {
    return Boolean(this.popup.link.length);
  }

  public hasPopup() {
    return this.hasPopupImage() || this.hasPopupTitle() || this.hasPopupText();
  }

  public getCategories() {
    if (!this.category || !this.category.length) {
      return;
    }

    return this.category.split(/[\r\n,]/g);
  }

  public hasCategory(category : string) {
    if (category.toLowerCase() === 'all') {
      return true;
    }

    let categories = this.getCategories();

    if (categories) {
      return categories.includes(category) || categories.includes('All') || categories.includes('all');
    }

    return false;
  }
};

class ImageSVGObject implements SVGObjectInterface {
  type : string = SVGObjectType.Image;
  category : string | undefined = undefined;
  position : Point = new Point();
  layer : number = 0;
  popup = {
    title: '',
    text: '',
    imageUrl: '',
    imageWidth : 0,
    imageHeight : 0,
    link: '',
  };

  width = 0;
  height = 0;
  scale = 1;
  imageUrl : string | undefined = undefined;

  constructor(o : object|undefined = undefined) {
    Object.assign(this, o);

    if (typeof o === 'object') {
      if ('position' in o && (typeof o.position === 'object') && o.position !== null && 'x' in o.position && 'y' in o.position && typeof o.position.x === 'number' && typeof o.position.y === 'number') {
        this.position = new Point(o.position.x, o.position.y);
      }

      this.layer = this.layer || 0;
    }
  }

  public getTransformOrigin() {
    return new Point(
      (this.position.x + this.width) / 2,
      (this.position.y + this.height) / 2,
    );
  }

  public getTransform() {
    return undefined;
  }
  
  public hasPopupImage() {
    return Boolean(this.popup.imageUrl && this.popup.imageUrl.length && this.popup.imageHeight && this.popup.imageWidth);
  }

  public hasPopupTitle() {
    return Boolean(this.popup.title.length);
  }

  public hasPopupText() {
    return Boolean(this.popup.text.length);
  }

  public hasPopupLink() {
    return Boolean(this.popup.link.length);
  }

  public hasPopup() {
    return this.hasPopupImage() || this.hasPopupTitle() || this.hasPopupText();
  }

  public getCategories() {
    if (!this.category || !this.category.length) {
      return;
    }

    return this.category.split(/[\r\n,]/g);
  }

  public hasCategory(category : string) {
    if (category.toLowerCase() === 'all') {
      return true;
    }

    let categories = this.getCategories();

    if (categories) {
      return categories.includes(category) || categories.includes('All') || categories.includes('all');
    }

    return false;
  }
};

class TextSVGObject implements SVGObjectInterface {
  type : string = SVGObjectType.Text;
  category : string | undefined = undefined;
  position : Point = new Point();
  layer : number = 0;
  popup = {
    title: '',
    text: '',
    imageUrl: '',
    imageWidth : 0,
    imageHeight : 0,
    link: '',
  };

  text : string | undefined = 'Text';
  textColor = '#000000FF';
  fontSize : number = 16;
  fontWeight : number = 400;
  fontStyle : string | null = null;

  fill = '#FFFFFF00';

  constructor(o : object|undefined = undefined) {
    Object.assign(this, o);

    if (typeof o === 'object') {
      if ('position' in o && (typeof o.position === 'object') && o.position !== null && 'x' in o.position && 'y' in o.position && typeof o.position.x === 'number' && typeof o.position.y === 'number') {
        this.position = new Point(o.position.x, o.position.y);
      }

      this.fontWeight = this.fontWeight || 400;
      this.fontStyle = this.fontStyle || null;
      this.layer = this.layer || 0;
    }
  }

  get textWidth () : number {
    if (this.text && this.text.length) {
      let rect = measureSVGText(this.text, this.fontSize + 'pt', undefined, this.fontWeight, this.fontStyle === 'italic' ? 'italic' : undefined, this.fontStyle === 'underline' ? 'underline' : undefined);

      if (rect) {
        return Math.max(rect.width, rect.height);
      }
    }

    return 0;
  }

  get textHeight () : number {
    if (this.text && this.text.length) {
      let rect = measureSVGText(this.text, this.fontSize + 'pt', undefined, this.fontWeight, this.fontStyle === 'italic' ? 'italic' : undefined, this.fontStyle === 'underline' ? 'underline' : undefined);

      if (rect) {
        return rect.height;
      }
    }

    return 0;
  }

  public getTransformOrigin() {
    return new Point(this.position.x, this.position.y);
  }

  public getTransform() {
    return undefined;
  }
  
  public hasPopupImage() {
    return Boolean(this.popup.imageUrl && this.popup.imageUrl.length && this.popup.imageHeight && this.popup.imageWidth);
  }

  public hasPopupTitle() {
    return Boolean(this.popup.title.length);
  }

  public hasPopupText() {
    return Boolean(this.popup.text.length);
  }

  public hasPopupLink() {
    return Boolean(this.popup.link.length);
  }

  public hasPopup() {
    return this.hasPopupImage() || this.hasPopupTitle() || this.hasPopupText();
  }

  public getCategories() {
    if (!this.category || !this.category.length) {
      return;
    }

    return this.category.split(/[\r\n,]/g);
  }

  public hasCategory(category : string) {
    if (category.toLowerCase() === 'all') {
      return true;
    }

    let categories = this.getCategories();

    if (categories) {
      return categories.includes(category) || categories.includes('All') || categories.includes('all');
    }

    return false;
  }
};

class LineSVGObject implements SVGObjectInterface {
  type : string = SVGObjectType.Line;
  category : string | undefined = undefined;
  position : Point = new Point();
  layer : number = 0;
  popup = {
    title: '',
    text: '',
    imageUrl: '',
    imageWidth : 0,
    imageHeight : 0,
    link: '',
  };

  stroke = '#000000FF';
  strokeWidth = 2;
  start : Point = new Point(-25, -25);
  end : Point = new Point(25, 25);
  pattern : string | undefined = 'dashed';
  arrowhead : boolean = false;

  constructor(o : object|undefined = undefined) {
    Object.assign(this, o);

    if (typeof o === 'object') {
      if ('position' in o && (typeof o.position === 'object') && o.position !== null && 'x' in o.position && 'y' in o.position && typeof o.position.x === 'number' && typeof o.position.y === 'number') {
        this.position = new Point(o.position.x, o.position.y);
      }

      if ('start' in o && (typeof o.start === 'object') && o.start !== null && 'x' in o.start && 'y' in o.start && typeof o.start.x === 'number' && typeof o.start.y === 'number') {
        this.start = new Point(o.start.x, o.start.y);
      }

      if ('end' in o && (typeof o.end === 'object') && o.end !== null && 'x' in o.end && 'y' in o.end && typeof o.end.x === 'number' && typeof o.end.y === 'number') {
        this.end = new Point(o.end.x, o.end.y);
      }

      this.layer = this.layer || 0;
    }
  }

  public getScaledDasharray() : string | undefined {
    if (this.pattern === 'dashed') {
      return [4, 2].map(v => v * this.strokeWidth).join(' ');
    }
    else if (this.pattern === 'dotted') {
      return [1, 1].map(v => v * this.strokeWidth).join(' ');
    }

    return undefined;
  }

  public getTransformOrigin() {
    return new Point(this.position.x + (this.start.x + this.end.x) / 2, this.position.y + (this.start.y + this.end.y) / 2);
  }

  public getTransform() {
    return undefined;
  }

  public hasPopupImage() {
    return Boolean(this.popup.imageUrl && this.popup.imageUrl.length && this.popup.imageHeight && this.popup.imageWidth);
  }

  public hasPopupTitle() {
    return Boolean(this.popup.title.length);
  }

  public hasPopupText() {
    return Boolean(this.popup.text.length);
  }

  public hasPopupLink() {
    return Boolean(this.popup.link.length);
  }

  public hasPopup() {
    return this.hasPopupImage() || this.hasPopupTitle() || this.hasPopupText();
  }

  public getCategories() {
    if (!this.category || !this.category.length) {
      return;
    }

    return this.category.split(/[\r\n,]/g).map(str => str.trim()).filter(str => str.length);
  }

  public hasCategory(category : string) {
    if (category.toLowerCase() === 'all') {
      return true;
    }

    let categories = this.getCategories();

    if (categories) {
      return categories.includes(category) || categories.includes('All') || categories.includes('all');
    }

    return false;
  }
};

class FilterSVGObject implements SVGObjectInterface {
  type : string = SVGObjectType.Filter;
  category : string | undefined = undefined;
  position : Point = new Point();
  layer : number = 0;
  popup = {
    title: '',
    text: '',
    imageUrl: '',
    imageWidth : 0,
    imageHeight : 0,
    link: '',
  };

  constructor(o : object|undefined = undefined) {
    Object.assign(this, o);

    if (typeof o === 'object') {
      if ('position' in o && (typeof o.position === 'object') && o.position !== null && 'x' in o.position && 'y' in o.position && typeof o.position.x === 'number' && typeof o.position.y === 'number') {
        this.position = new Point(o.position.x, o.position.y);
      }

      this.layer = this.layer || 0;
    }
  }

  public getTransformOrigin() {
    return new Point(this.position.x, this.position.y);
  }

  public getTransform() {
    return undefined;
  }

  public hasPopupImage() {
    return Boolean(this.popup.imageUrl && this.popup.imageUrl.length && this.popup.imageHeight && this.popup.imageWidth);
  }

  public hasPopupTitle() {
    return Boolean(this.popup.title.length);
  }

  public hasPopupText() {
    return Boolean(this.popup.text.length);
  }

  public hasPopupLink() {
    return Boolean(this.popup.link.length);
  }

  public hasPopup() {
    return this.hasPopupImage() || this.hasPopupTitle() || this.hasPopupText();
  }

  public getCategories() {
    if (!this.category || !this.category.length) {
      return;
    }

    return this.category.split(/[\r\n,]/g).map(str => str.trim()).filter(str => str.length);
  }

  public hasCategory(category : string) {
    return category ? true : true;
  }
};

function isSVGObject(o : any | SVGObjectInterface) : o is SVGObjectInterface {
  return typeof o === 'object' && (o instanceof AreaSVGObject || o instanceof ImageSVGObject || o instanceof TextSVGObject || o instanceof LineSVGObject || o instanceof FilterSVGObject);
}

function upgradeStorageClasses (storageData : StorageInterface) : StorageInterface {
  storageData.elements = storageData.elements || [];
  storageData.elements = storageData.elements.map((v : SVGObjectInterface) => {
    switch (v.type) {
      case SVGObjectType.Area:
        return new AreaSVGObject(v);

      case SVGObjectType.Image:
        return new ImageSVGObject(v);

      case SVGObjectType.Text:
        return new TextSVGObject(v);
      
      case SVGObjectType.Line:
        return new LineSVGObject(v);

      case SVGObjectType.Filter:
        return new FilterSVGObject(v);

      default:
        throw new Error('Invalid object!');
    }
  });

  return storageData;
}

function getSVGPos(svg: SVGGraphicsElement | undefined, clientX : number | MouseEvent, clientY : number | undefined = undefined) : Point {
  if (clientX instanceof MouseEvent) {
    return getSVGPos(svg, clientX.clientX, clientX.clientY);
  }

  let ret = new Point();

  if (clientY === undefined) {
    return ret;
  }

  if (svg instanceof SVGGraphicsElement) {
    let ctm = svg.getScreenCTM();

    if (ctm) {
      ret.x = (clientX - ctm.e) / ctm.a;
      ret.y = (clientY - ctm.f) / ctm.d;
    }
  }

  return ret;
}

function generateCrossHatch(width : number, height : number, d : number = 6) : string[] {
  let ret : string[] = [], n = Math.max(width, height) + d;

  for (let i = -n * 10 + d / 2; i <= n * 10; i += d) {
    ret.push([
      [i + n - 1, i - n - 1].join(','),
      [i + n + 1, i - n + 1].join(','),
      [i - n + 1, i + n + 1].join(','),
      [i - n - 1, i + n - 1].join(','),
    ].join(' '));
  }

  return ret;
}

function getAllCategories(s : StorageInterface) : string[] {
  let dict : { [name : string] : string } = {};

  s.elements.forEach(elem => {
    let categories = elem.getCategories();

    if (categories) {
      categories.forEach(c => {
        dict[c] = c;
      });
    }
  });

  return ['All', ...Object.values(dict).filter(str => str.toLocaleLowerCase() !== 'all').sort()];
}

function getLayers(elements : SVGObjectInterface[]) {
  let ret : {
    [key : string]: number,
  } = {};

  elements.forEach(elem => {
    ret[elem.layer] = elem.layer;
  });

  return Object.values(ret).sort((a, b) => a - b);
}

export {
  SVGObjectType,
  appBase,
  publicUrl,
  NumberPair,
  Point,
  AreaSVGObject,
  ImageSVGObject,
  TextSVGObject,
  LineSVGObject,
  FilterSVGObject,
  isSVGObject,
  upgradeStorageClasses,
  getSVGPos,
  generateCrossHatch,
  measureSVGText,
  getAllCategories,
  reformatHTML,
  getLayers,
};

export default {
  SVGObjectType,
  appBase,
  publicUrl,
  NumberPair,
  Point,
  AreaSVGObject,
  ImageSVGObject,
  TextSVGObject,
  LineSVGObject,
  FilterSVGObject,
  isSVGObject,
  upgradeStorageClasses,
  getSVGPos,
  generateCrossHatch,
  measureSVGText,
  getAllCategories,
  reformatHTML,
  getLayers,
};
